/*
* 
* SPDX-License-Identifier: PROPRIETARY
*
* (C)Copyright 2023, Yellowtec GmbH
* author: David Zimmermann (dZim), dzimmermann@yellowtec.com
*
*/

#ifndef LIB_HUSH_H
#define LIB_HUSH_H

#include <wchar.h>

#ifdef _WIN32
      #define LHUSH_EXPORT __declspec(dllexport)
      #define LHUSH_CALL
#else
      #define LHUSH_EXPORT
      #define LHUSH_CALL
#endif

#ifdef __cplusplus
extern "C" {
#endif


//*****************************************************************************
// Definition

//! Max device name string size (including trailing end mark)
#define MAX_SIZE_DEVICE_NAME    (40+1)
//! Max serial number string size (including trailing end mark)
#define MAX_SIZE_SERIAL_NUMBER  (10+1)

//! Version number nibble
typedef unsigned int lhush_version_nibble_t;

//! These result codes will be used as indication of success or failure.
typedef enum
{
    LHUSH_RC_SUCCESS = 0,                           /*!< successful completion */
    //
    LHUSH_RC_ERROR_INVALID_ARGUMENT,                /*!< one or more function arguments do not meet expectations */
    LHUSH_RC_ERROR_DEVICE_NOT_AVAILABLE,            /*!< the requested hush HID device is not available on this host */
    LHUSH_RC_ERROR_OUT_OF_MEMORY,                   /*!< allocation of memory failed */
    LHUSH_RC_ERROR_CANT_WRITE,                      /*!< cannot write to hush HID device */
    LHUSH_RC_ERROR_CANT_READ,                       /*!< cannot read from hush HID device */
    LHUSH_RC_ERROR_INTERNAL                         /*!< internal error */

} lhush_result_codes_t;

/*!
    \brief Structure containing device information.
*/
typedef struct lhush_device_info_t
{
    wchar_t device_name[MAX_SIZE_DEVICE_NAME];      /*!< hush device name */
    wchar_t serial_number[MAX_SIZE_SERIAL_NUMBER];  /*!< hush device serial number */

} lhush_device_info_t;

/*!
    \brief Enumeration to indicate hush operation modes.

    A hush device is in one operation mode at a time.
    The operation mode defines which signal sources will be evaluated to alter device's operation state.
*/
typedef enum
{
    LHUSH_MD_REMOTE = 0,                            /*!< remote control */
    LHUSH_MD_GPI,                                   /*!< GPI */
    LHUSH_MD_USB_HID,                               /*!< USB HID */
    LHUSH_MD_USB_MIDI,                              /*!< USB MIDI */
    LHUSH_MD_MIDI,                                  /*!< UART MIDI */
    //
    LHUSH_MD_UNKNOWN

} lhush_mode_t;

/*!
    \brief Bit field to define hush operation states.

    A hush device might have none, one or more states enabled at the same time.
    An enabled red state has precedence over enabled white state.
    Mute enable has an effect on hush+ devices only with an preamp included.
*/
typedef struct
{
   unsigned char white_enable : 1;                  /*!< white m!ka LED enable */
   unsigned char red_enable : 1;                    /*!< red m!ka LED enable */
   unsigned char mute_enable : 1;                   /*!< preamp mute enable */

} lhush_state_t;


//*****************************************************************************
// Interface

/*!
    \brief Call this function to get version information for hush library.

    It returns all three nibbles of version number information.

    \param p_version_major   pointer to major version number nibble
    \param p_version_minor   pointer to minor version number nibble
    \param p_version_patch   pointer to patch version number nibble
    \return                  returns result code for the getting version
*/
LHUSH_EXPORT lhush_result_codes_t LHUSH_CALL lhush_get_version_info(
    lhush_version_nibble_t * p_version_major,
    lhush_version_nibble_t * p_version_minor,
    lhush_version_nibble_t * p_version_patch);

/*!
    \brief Call this function to acquire name and serial number information for all connected hush devices.

    It returns an array of lhush_device_info_t structures containing both device name and device serial number.
    Serial numbers might be used to get/set state from/to defined devices in a multi device setup.
    The acquired device info array has to be released via lhush_release_device_infos() afterwards.

    \param pp_device_infos      pointer to pointer to first device info of array
    \param p_num_device_infos   pointer to number of elements of array
    \return                     returns the overall result code for the enumeration procedure
*/
LHUSH_EXPORT lhush_result_codes_t LHUSH_CALL lhush_acquire_device_infos(lhush_device_info_t ** pp_device_infos, size_t * p_num_device_infos);

/*!
    \brief Call this function to release device info array, previously aqcuired via lhush_acquire_device_infos()

    \param pp_device_infos      pointer to pointer to first device info (items are single linked via 'p_next')
*/
LHUSH_EXPORT void LHUSH_CALL lhush_release_device_infos(lhush_device_info_t ** pp_device_infos);

/*!
    \brief Call this function to set operation state on hush device.

    This function alters operation state and in case also  operation mode of hush device.
    Only masked states will be touched while others remain 'as is'.
    The device will implicitly forced to enter 'USB HID' operation mode.
    If no serial number will be given (NULL) the very first device found on host will be used, which is a valid shortcut on single device setups.
    If a serial number will be given write operation will be tried on this specified device only (approach for multi device setups).
    On a given serial number, operation fails if no such device is connected even though there might be other hush devices available on the same host.

    \param serial_number        the serial number of the device to set state on (or NULL)
    \param mask                 masks which operation states shall be touched
    \param state                defines operation states
    \return                     returns result code
*/
LHUSH_EXPORT lhush_result_codes_t LHUSH_CALL lhush_set_state(wchar_t const * serial_number, lhush_state_t mask, lhush_state_t state);

//! 
/*!
    \brief Call this function to get operation mode & state on hush device.

    This function does not alter operation state or operation mode of hush device.
    It might be used to monitor device's operation mode and state and does not force the device to be in USB HID mode.
    If no serial number will be given (NULL) the very first device found on host will be used, which is a valid shortcut on single device setups.
    If a serial number will be given read operation will be tried on this specified device only (approach for multi device setups).
    On a given serial number, operation fails if no such device is connected even though there might be other hush devices available on the same host.

    \param serial_number        the serial number of the device to get mode & state from (or NULL)
    \param p_mode               points to operation mode
    \param p_state              points to operation state
    \return                     returns result code
*/
LHUSH_EXPORT lhush_result_codes_t LHUSH_CALL lhush_get_mode_and_state(wchar_t const * serial_number, lhush_mode_t* p_mode, lhush_state_t* p_state);

#ifdef __cplusplus
}
#endif

#endif // LIB_HUSH_H