/*
* 
* SPDX-License-Identifier: MIT
*
* (C)Copyright 2023, Yellowtec GmbH
* author: David Zimmermann (dZim), dzimmermann@yellowtec.com
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <wchar.h>
//
#include "libhush.h"
#include "config.h"

//*****************************************************************************
// Definition

#define TOOLNAME            "toolhush"

#define ACTION_LIST         "list"
#define ACTION_SET          "set"
#define ACTION_GET          "get"

#define STATE_WHITE         'w'
#define STATE_RED           'r'
#define STATE_MUTE          'm'

#define STRINGIFY_(EXP)     #EXP
#define STRINGIFY(EXP)      STRINGIFY_(EXP)

#define MIN(_x, _y)         (((_x) < (_y)) ? (_x) : (_y))


//*****************************************************************************
// Internal

// generic banner message
static void banner_message()
{
    lhush_version_nibble_t version_major = 0;
    lhush_version_nibble_t version_minor = 0;
    lhush_version_nibble_t version_patch = 0;

    lhush_get_version_info(&version_major, &version_minor, &version_patch);

    fprintf(    stdout,
                "\nlibhush v%d.%d.%d\n%s v%d.%d.%d\n\n"
                "Copyright (c) 2023 Yellowtec GmbH\n\n"
                "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,\n" 
                "INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.\n" 
                "IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,\n"
                "WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n" 
                "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n\n",
                version_major, version_minor, version_patch,
                TOOLNAME, THUSH_VERSION_MAJOR, THUSH_VERSION_MINOR, THUSH_VERSION_PATCH);
}

static void stop_message(const char* p_message, char success)
{
    fprintf(    success ? stdout : stderr,
                "%s -> %s\n",
                success ? "SUCCESS" : "ERROR", p_message);
}

// printout what was wrong, how to use the tool and leave
static void exit_with_usage(const char* p_tool_name, const char* p_message)
{
    stop_message(p_message, 0);
    printf("USAGE: The tool supports three different actions: '" ACTION_LIST "', '" ACTION_GET "' and '" ACTION_SET "'.\n");
    printf("\tOn '" ACTION_LIST "' it lists all connected hush(+) devices with serial number and name.\n");
    printf("\tOn '" ACTION_GET "' it gets current operation mode and states from a given device.\n");
    printf("\tOn '" ACTION_SET "' it sets operation states on a given device (operation mode will be set impilictly to 'USB HID')\n"); 
    printf("\n\t'" ACTION_LIST "' needs no further arguments:\n");
    printf("\t\tEXAMPLE: %s " ACTION_LIST "\n", p_tool_name);
    printf("\n\t'" ACTION_GET "' supports 1 optional argument:\n");
    printf("\t\t1st: serial number of device to request mode and state on (nothing means very first device found on host)\n");
    printf("\t\tEXAMPLE1: %s " ACTION_GET " 23HS000012\n", p_tool_name);
    printf("\t\tEXAMPLE2: %s " ACTION_GET "\n", p_tool_name);
    printf("\n\t'" ACTION_SET "' supports 2 optional arguments:\n");
    printf("\t\t1st: states to enable as character codes ("" means all off)\n");
    printf("\t\t2nd: serial number of the device to set state on (nothing means very first device found on host).\n");
    printf("\t\t\t- '%c' to enable white m!ka LED\n", STATE_WHITE);
    printf("\t\t\t- '%c' to enable red m!ka LED\n", STATE_RED);
    printf("\t\t\t- '%c' to enable mute (supported on hush+ only)\n", STATE_MUTE);
    printf("\t\tEXAMPLE1 (enable white & mute, disable red): %s " ACTION_SET " \"wm\" 23HS000012\n", p_tool_name);
    printf("\t\tEXAMPLE2 (enable red, first found device): %s " ACTION_SET " \"r\"\n", p_tool_name);
    printf("\t\tEXAMPLE3 (all off, first found device): %s " ACTION_SET " \"\"\n", p_tool_name);
    printf("\t\tEXAMPLE4 (all off): %s " ACTION_SET " \"\" 23HS000012\n", p_tool_name);
    exit(1);
}

// printout what was wrong, how to use the tool and leave
static void exit_action_with_error(const char* p_message, wchar_t const * wserial, lhush_result_codes_t res)
{
    stop_message(p_message, 0);
    if (wserial)
    {
        fprintf(stderr,
                "libhush res -> '%d' for device '%ls'\n",
                res, wserial);
    }
    else
    {
        fprintf(stderr,
                "libhush res -> '%d'\n",
                res);
    }
    exit(2);
}

static wchar_t const * aserial_to_wserial(char const * p_aserial)
{
    static wchar_t  wserial[11];
    int             i_char;

    // convert ascii serial to wchar
    for (i_char = 0; i_char < MIN(10, strlen(p_aserial)); i_char++)
    {
        wserial[i_char] = (wchar_t)p_aserial[i_char]; 
    }

    // enforce termination
    wserial[i_char] = L'\0';

    return wserial;
}

static lhush_state_t string_to_mask(char const * p_amask)
{
    lhush_state_t state = 
    {
        .white_enable   = (NULL != strchr(p_amask, STATE_WHITE))    || (NULL != strchr(p_amask, toupper(STATE_WHITE))),
        .red_enable     = (NULL != strchr(p_amask, STATE_RED))      || (NULL != strchr(p_amask, toupper(STATE_RED))),
        .mute_enable    = (NULL != strchr(p_amask, STATE_MUTE))     || (NULL != strchr(p_amask, toupper(STATE_MUTE)))
    };

    return state;
}

static void print_state(lhush_state_t state)
{
    unsigned int num_elements = 0;

    if (state.mute_enable)
    {
        printf("MUTE");
        num_elements++;
    }
    if (state.red_enable)
    {
        if (num_elements)
        {
            printf(" & ");
        }
        printf("RED");
        num_elements++;
    }
    if (state.white_enable)
    {
        if (state.red_enable)
        {
            printf(" (RED overrides WHITE)");
        }
        else
        {
            if (num_elements)
            {
                printf(" & ");
            }
            printf("WHITE");
            num_elements++;
        }
    }

    printf((num_elements) ? "\n":"ALL OFF\n");
}

char const * mode_to_text(mode_to_get)
{
    switch (mode_to_get)
    {
        case LHUSH_MD_REMOTE:
            return "REMOTE";
        case LHUSH_MD_GPI:
            return "GPI";
        case LHUSH_MD_USB_HID:
            return "USB HID";
        case LHUSH_MD_USB_MIDI:
            return "USB MIDI";
        case LHUSH_MD_MIDI:
            return "MIDI";
    }

    return "UNKNWON";
}


//*****************************************************************************
// Interface

int main(int argc, char **argv)
{
    char const * p_tool_name;
    char const * p_action;

    banner_message();

    p_tool_name = argv[0];

    // sufficient arguments ?
    // - we expect product name and action as both arguments
    if (argc < 2)
    {
        exit_with_usage(p_tool_name, "Insufficient arguments.");
    }

    p_action = argv[1];

    if (0 == strcmp(p_action, ACTION_LIST))
    {
        lhush_device_info_t * p_device_info;
        size_t num_device_infos;
        lhush_result_codes_t res = lhush_acquire_device_infos(&p_device_info, &num_device_infos);

        if (LHUSH_RC_SUCCESS != res)
        {
            exit_action_with_error("listing devices acquiring device infos", NULL, res);
        }

        if ((NULL == p_device_info) || (0 == num_device_infos))
        {
            printf("\nNo Devices found!\n");
        }
        else
        {
            printf("\nDevices found:\n");

            for (size_t i_dev = 0; i_dev < num_device_infos; i_dev++)
            {
                printf("\t[%zu]\tserial: '%ls' -> name: '%ls'\n", i_dev, p_device_info[i_dev].serial_number, p_device_info[i_dev].device_name);
            }
 
            printf("\n");
        }

        lhush_release_device_infos(&p_device_info);

        printf("\nlisting devices, done.\n");
    }
    else if (0 == strcmp(p_action, ACTION_SET))
    {
        // harvest arguments
        lhush_state_t state_to_set  = string_to_mask((argc >= 3) ? argv[2] : "");
        wchar_t const * p_wserial   = (argc >= 4) ? aserial_to_wserial(argv[3]) : NULL;

        printf("\nset: ");
        print_state(state_to_set);

        // control all states at once
        lhush_state_t mask =
        {
            .white_enable   = 1,
            .red_enable     = 1,
            .mute_enable    = 1
        };

        lhush_result_codes_t res = lhush_set_state(p_wserial, mask, state_to_set);

        if (LHUSH_RC_SUCCESS != res)
        {
            exit_action_with_error("set state failed on set state", p_wserial, res);
        }

        printf("\nsetting state, done.\n");
    }
    else if (0 == strcmp(p_action, ACTION_GET))
    {
        // harvest arguments
        wchar_t const * p_wserial = (argc >= 3) ? aserial_to_wserial(argv[2]) : NULL;

        lhush_mode_t  mode_to_get;
        lhush_state_t state_to_get;
        lhush_result_codes_t res = lhush_get_mode_and_state(p_wserial, &mode_to_get, &state_to_get);

        if (LHUSH_RC_SUCCESS != res)
        {
            exit_action_with_error("get mode & state failed on get state", p_wserial, res);
        }
        else
        {
            printf("\nmode '%s', state: ", mode_to_text(mode_to_get));
            print_state(state_to_get);
        }

        printf("\ngetting mode & state, done.\n");
    }
    else
    {
        exit_with_usage(p_tool_name, "Unsupported action.");
    }
 
    return 0;
}
